from typing import Optional

from cpl_core.database.context import DatabaseContextABC
from cpl_query.extension import List

from bot_api.filter.discord.server_select_criteria import ServerSelectCriteria
from bot_core.logging.database_logger import DatabaseLogger
from bot_data.abc.server_repository_abc import ServerRepositoryABC
from bot_data.filtered_result import FilteredResult
from bot_data.model.server import Server
from bot_data.service.cache_service import CacheService


class ServerRepositoryService(ServerRepositoryABC):
    def __init__(self, logger: DatabaseLogger, db_context: DatabaseContextABC, cache: CacheService):
        self._logger = logger
        self._context = db_context
        self._cache = cache

        ServerRepositoryABC.__init__(self)

    def get_servers(self) -> List[Server]:
        servers = List(Server)
        self._logger.trace(__name__, f"Send SQL command: {Server.get_select_all_string()}")
        results = self._context.select(Server.get_select_all_string())
        for result in results:
            servers.append(Server(result[1], result[2], result[3], id=result[0]))

        self._cache.cached_server = List(Server, servers)
        return servers

    def get_filtered_servers(self, criteria: ServerSelectCriteria) -> FilteredResult:
        servers = self.get_servers()
        self._logger.trace(__name__, f"Send SQL command: {Server.get_select_all_string()}")
        query = servers

        # sort
        if (
            criteria.sort_column is not None
            and criteria.sort_column != ""
            and criteria.sort_direction is not None
            and criteria.sort_direction
        ):
            crit_sort_direction = criteria.sort_direction.lower()
            if crit_sort_direction == "desc" or crit_sort_direction == "descending":
                query = query.order_by_descending(lambda x: getattr(x, criteria.sort_column))
            else:
                query = query.order_by(lambda x: getattr(x, criteria.sort_column))

        result = FilteredResult()
        result.total_count = query.count()
        skip = criteria.page_size * criteria.page_index
        result.result = query.skip(skip).take(criteria.page_size)

        return result

    def get_server_by_id(self, server_id: int) -> Server:
        cs = self._cache.cached_server.where(lambda x: x.id == server_id).single_or_default()
        if cs is not None:
            return cs

        self._logger.trace(__name__, f"Send SQL command: {Server.get_select_by_id_string(server_id)}")
        result = self._context.select(Server.get_select_by_id_string(server_id))[0]
        server = Server(result[1], result[2], result[3], id=result[0])
        self._cache.cached_server.add(server)
        return server

    def get_server_by_discord_id(self, discord_id: int) -> Server:
        cs = self._cache.cached_server.where(lambda x: x.discord_id == discord_id).single_or_default()
        if cs is not None:
            return cs

        self._logger.trace(
            __name__,
            f"Send SQL command: {Server.get_select_by_discord_id_string(discord_id)}",
        )
        result = self._context.select(Server.get_select_by_discord_id_string(discord_id))[0]
        return Server(result[1], result[2], result[3], id=result[0])

    def find_server_by_discord_id(self, discord_id: int) -> Optional[Server]:
        cs = self._cache.cached_server.where(lambda x: x.discord_id == discord_id).single_or_default()
        if cs is not None:
            return cs

        self._logger.trace(
            __name__,
            f"Send SQL command: {Server.get_select_by_discord_id_string(discord_id)}",
        )
        result = self._context.select(Server.get_select_by_discord_id_string(discord_id))
        if result is None or len(result) == 0:
            return None

        result = result[0]

        return Server(result[1], result[2], result[3], id=result[0])

    def add_server(self, server: Server):
        self._logger.trace(__name__, f"Send SQL command: {server.insert_string}")
        self._context.cursor.execute(server.insert_string)

    def update_server(self, server: Server):
        self._logger.trace(__name__, f"Send SQL command: {server.udpate_string}")
        self._context.cursor.execute(server.udpate_string)

    def delete_server(self, server: Server):
        self._logger.trace(__name__, f"Send SQL command: {server.delete_string}")
        self._context.cursor.execute(server.delete_string)
