import discord
from cpl_core.configuration import ConfigurationABC
from cpl_core.database.context import DatabaseContextABC
from cpl_core.logging import LoggerABC
from cpl_discord.container import Guild, Role, Member
from cpl_discord.service import DiscordBotServiceABC
from cpl_translation import TranslatePipe

from bot_core.service.message_service import MessageService
from bot_data.model.level import Level
from bot_data.model.user import User
from bot_data.service.level_repository_service import LevelRepositoryService
from bot_data.service.server_repository_service import ServerRepositoryService
from bot_data.service.user_repository_service import UserRepositoryService
from modules.level.configuration.level_server_settings import LevelServerSettings


class LevelService:
    def __init__(
        self,
        config: ConfigurationABC,
        logger: LoggerABC,
        db: DatabaseContextABC,
        levels: LevelRepositoryService,
        users: UserRepositoryService,
        servers: ServerRepositoryService,
        bot: DiscordBotServiceABC,
        message_service: MessageService,
        t: TranslatePipe,
    ):
        self._config = config
        self._logger = logger
        self._db = db
        self._levels = levels
        self._users = users
        self._servers = servers
        self._bot = bot
        self._message_service = message_service
        self._t = t

    def get_level(self, user: User) -> Level:
        levels_by_server = self._levels.get_levels_by_server_id(user.server.id)
        levels = levels_by_server.order_by(lambda l: l.min_xp).where(lambda l: user.xp >= l.min_xp)

        if levels.count() == 0:
            return levels_by_server.order_by(lambda l: l.min_xp).last()

        return levels.last()

    async def set_level(self, user: User):
        level_names = self._levels.get_levels_by_server_id(user.server.id).select(lambda l: l.name)
        guild: Guild = self._bot.guilds.where(lambda g: g.id == user.server.discord_id).single()
        member: Member = guild.members.where(lambda m: m.id == user.discord_id).single()

        level = self.get_level(user)
        level_role: Role = guild.roles.where(lambda r: r.name == level.name).single()
        if level_role in member.roles:
            return

        notification_needed = False
        for role in member.roles.where(lambda r: r.name in level_names.to_list()):
            try:
                self._logger.debug(__name__, f"Try to remove role {role.name} from {member.name}")
                await member.remove_roles(role)
                notification_needed = True
                self._logger.info(__name__, f"Removed role {role.name} from {member.name}")
            except Exception as e:
                self._logger.error(__name__, f"Removing role {role.name} from {member.name} failed!", e)

        try:
            self._logger.debug(__name__, f"Try to add role {level_role.name} to {member.name}")
            await member.add_roles(level_role)
            self._logger.info(__name__, f"Add role {level_role.name} to {member.name}")
        except Exception as e:
            self._logger.error(__name__, f"Adding role {level_role.name} to {member.name} failed!", e)

        if notification_needed:
            level_settings: LevelServerSettings = self._config.get_configuration(f"LevelServerSettings_{guild.id}")
            await self._message_service.send_channel_message(
                self._bot.get_channel(level_settings.changed_level_notification_channel),
                self._t.transform("modules.level.new_level_message").format(member.mention, level.name),
                is_persistent=True,
            )

    async def check_level(self, member: discord.Member):
        if member.bot:
            return

        server = self._servers.get_server_by_discord_id(member.guild.id)
        user = self._users.find_user_by_discord_id_and_server_id(member.id, server.id)
        if user is None:
            self._logger.warn(__name__, f"User not found {member.guild.name}@{member.name}")
            return

        await self.set_level(user)
